from pydantic import BaseModel, Field, RootModel


class CorrelationCandidate(BaseModel):
    id: str = Field(..., description="Unique identifier for the candidate")
    x_centroid: float = Field(..., description="X coordinate of the streak centroid in pixels")
    y_centroid: float = Field(..., description="Y coordinate of the streak centroid in pixels")
    range: float = Field(..., description="Estimated range to the object in kilometers")


class CorrelationCandidates(RootModel[list[CorrelationCandidate]]):

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, index: int) -> CorrelationCandidate:
        return self.root[index]

    def __iter__(self):
        return iter(self.root)



from pathlib import Path
from typing import Any
from urllib.request import urlretrieve

import numpy as np
import numpy.typing as npt
import tetra3  # type: ignore
from pydantic import BaseModel, Field

from pixelemon.constants import ZERO
from pixelemon.logging import PixeLemonLog

_GH_URL = "https://github.com/citra-space/pixelemon/releases/download/general-purpose-database"
TETRA_DB_NAME = "tyc_db_to_40_deg.npz"
TETRA_DATABASE_URL = f"{_GH_URL}/{TETRA_DB_NAME}"
TETRA_DATABASE_PATH = Path(__file__).parent / TETRA_DB_NAME


class TetraSolver(tetra3.Tetra3):

    _instance: "TetraSolver | None" = None
    _initialized: bool = False

    def __new__(cls, *args, **kwargs) -> "TetraSolver":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, database_path: Path | None = None, reload: bool = False):
        if not self._initialized or reload:
            if database_path is None:
                super().__init__()
            else:
                super().__init__(load_database=database_path.as_posix())
            self.settings = TetraSettings.model_validate(self.database_properties)
            self._initialized = True

    @classmethod
    def low_memory(cls) -> "TetraSolver":
        solver = cls(reload=True)
        return solver

    @classmethod
    def high_memory(cls) -> "TetraSolver":
        if not TETRA_DATABASE_PATH.exists():
            PixeLemonLog().warning("Tetra database not found locally.  One-time download may take several minutes.")
            PixeLemonLog().info(f"Downloading Tetra database from {_GH_URL} to {TETRA_DATABASE_PATH}")
            urlretrieve(TETRA_DATABASE_URL, TETRA_DATABASE_PATH.as_posix())
        solver = cls(TETRA_DATABASE_PATH, reload=True)
        return solver

    def solve_from_centroids(
        self,
        star_centroids: npt.NDArray[np.float32],
        img_size: tuple[int, int],
        fov: float,
    ) -> "PlateSolve | None":
        try:
            result = super().solve_from_centroids(star_centroids, size=img_size, fov_estimate=fov, return_matches=True)
            solve = PlateSolve(result)
            PixeLemonLog().info(f"Completed plate solve in {solve.solve_time:.2f} ms")
        except ValueError as e:
            PixeLemonLog().error(e)
            return None
        return solve


class TetraSettings(BaseModel):
    dimmest_star_magnitude: float = Field(
        ...,
        description="The dimmest star magnitude to consider in the plate solve",
        alias="star_max_magnitude",
    )
    max_field_of_view: float = Field(
        ...,
        description="The maximum field of view in degrees the solver can reliably handle",
        alias="max_fov",
    )
    min_field_of_view: float = Field(
        ...,
        description="The minimum field of view in degrees the solver can reliably handle",
        alias="min_fov",
    )
    verification_star_count: int = Field(
        ...,
        description="The number of stars to use for verification of the plate solve",
        alias="verification_stars_per_fov",
    )

    def model_post_init(self, _: Any) -> None:
        PixeLemonLog().info(f"Tetra dimmest star magnitude set to {self.dimmest_star_magnitude}")
        PixeLemonLog().info(f"Tetra max field of view set to {self.max_field_of_view} degrees")
        PixeLemonLog().info(f"Tetra min field of view set to {self.min_field_of_view} degrees")
        PixeLemonLog().info(f"Tetra verification star count set to {self.verification_star_count}")


class PlateSolve:

    def __init__(self, tetra_output):
        if tetra_output is None:
            raise ValueError("Plate solve failed")
        elif tetra_output["Prob"] > ZERO:
            raise ValueError("Plate solve returned false positive probability greater than zero")

        self.right_ascension = tetra_output["RA"]
        self.declination = tetra_output["Dec"]
        self.roll = tetra_output["Roll"]
        self.estimated_horizontal_fov = tetra_output["FOV"]
        self.root_mean_square_error = tetra_output["RMSE"]
        self.number_of_stars = tetra_output["Matches"]
        self.false_positive_probability = tetra_output["Prob"]
        self.solve_time = tetra_output["T_solve"]
        self.matched_stars = np.array(tetra_output["matched_stars"])
        self.matched_centroids = np.array(tetra_output["matched_centroids"])
        self.distortion = tetra_output["distortion"]



MICRO_TO_MILLI = 1e-3
MILLI_TO_MICRO = 1e3
DEGREES_TO_SECONDS = 3600.0
SECONDS_TO_DEGREES = 1 / 3600.0
INCHES_TO_MILLIMETERS = 25.4
PERCENT_TO_DECIMAL = 0.01
FWHM_TO_SIGMA = 1 / 2.355
BASE_TO_KILO = 1e-3
KILO_TO_BASE = 1e3
SUN_MAGNITUDE = -26.74
ZERO = 1e-12
MAD_TO_STD = 1.4826



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel, RootModel


class PixeLemonObservation(BaseModel):

    satellite_id: UUID
    epoch: datetime
    right_ascension: float
    declination: float
    sensor_latitude: float
    sensor_longitude: float
    sensor_altitude: float
    angular_noise: float
    visual_magnitude: float


class PixeLemonObservations(RootModel[list[PixeLemonObservation]]):

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, index: int) -> PixeLemonObservation:
        return self.root[index]

    def __iter__(self):
        return iter(self.root)



from pydantic import Field

from pixelemon.sensors._base_sensor import BaseSensor


class IMX174(BaseSensor):
    x_pixel_count: int = Field(default=1920, frozen=True)
    y_pixel_count: int = Field(default=1200, frozen=True)
    pixel_width: float = Field(default=5.86, frozen=True)
    pixel_height: float = Field(default=5.86, frozen=True)



import numpy as np
import numpy.typing as npt

from pixelemon.constants import MAD_TO_STD


def get_median_absolute_deviation(data: npt.NDArray[np.float32]) -> float:
    median = np.median(data)
    return float(np.median(np.abs(data - median)))


def get_standard_deviation(data: npt.NDArray[np.float32], use_mad: bool) -> float:
    if use_mad:
        mad = get_median_absolute_deviation(data)
        return mad * MAD_TO_STD
    else:
        return float(np.std(data))


def map_within_std_threshold(dat: npt.NDArray[np.float32], lim: float, use_mad: bool) -> npt.NDArray[np.bool_]:
    """Returns a boolean mask where True indicates the value is within the specified standard deviation threshold."""
    lower, upper = get_std_threshold_bounds(dat, lim, use_mad)
    return (dat >= lower) & (dat <= upper)


def map_outside_std_threshold(dat: npt.NDArray[np.float32], lim: float, use_mad: bool) -> npt.NDArray[np.bool_]:
    """Returns a boolean mask where True indicates the value is outside the specified standard deviation threshold."""
    lower, upper = get_std_threshold_bounds(dat, lim, use_mad)
    return (dat < lower) | (dat > upper)


def get_std_threshold_bounds(dat: npt.NDArray[np.float32], lim: float, use_mad: bool) -> tuple[float, float]:
    """Returns the lower and upper bounds based on the specified standard deviation threshold."""
    std = get_standard_deviation(dat, use_mad)
    if use_mad:
        center = np.median(dat)
    else:
        center = np.mean(dat)
    lower_bound = center - lim * std
    upper_bound = center + lim * std
    return float(lower_bound), float(upper_bound)


def map_outside_value_threshold(dat: npt.NDArray[np.float32], min_val: float, max_val: float) -> npt.NDArray[np.bool_]:
    """Returns a boolean mask where True indicates the value is outside the specified value thresholds."""
    return (dat < min_val) | (dat > max_val)



import numpy as np
from numpy import typing as npt
from pydantic import BaseModel, Field, RootModel

from pixelemon.constants import MAD_TO_STD


class DetectionLine(BaseModel):
    slope: float = Field(..., description="Slope of the streak line")
    intercept: float = Field(..., description="Y-intercept of the streak line")
    length: float = Field(..., description="Length of the streak line in pixels")
    x_centroid: float = Field(..., description="X coordinate of the streak centroid in pixels")
    y_centroid: float = Field(..., description="Y coordinate of the streak centroid in pixels")

    def get_distance(self, x: float, y: float) -> float:
        """Calculate the perpendicular distance from a point to the streak line."""
        numerator = abs(self.slope * x - y + self.intercept)
        denominator = np.sqrt(self.slope**2 + 1)
        return numerator / denominator

    @property
    def x0(self) -> float:
        """Get the starting x coordinate of the streak line."""
        return self.x_centroid - (self.length / 2) / np.sqrt(1 + self.slope**2)

    @property
    def y0(self) -> float:
        """Get the starting y coordinate of the streak line."""
        return self.slope * self.x0 + self.intercept

    @property
    def x1(self) -> float:
        """Get the ending x coordinate of the streak line."""
        return self.x_centroid + (self.length / 2) / np.sqrt(1 + self.slope**2)

    @property
    def y1(self) -> float:
        """Get the ending y coordinate of the streak line."""
        return self.slope * self.x1 + self.intercept


class Detection(BaseModel):
    x_centroid: float = Field(..., description="X coordinate of the detection centroid in pixels")
    y_centroid: float = Field(..., description="Y coordinate of the detection centroid in pixels")
    semi_major_axis: float = Field(..., description="Semi-major axis of the detection in pixels")
    semi_minor_axis: float = Field(..., description="Semi-minor axis of the detection in pixels")
    angle_to_horizon: float = Field(..., description="Orientation angle of the detection in radians")
    total_flux: float = Field(..., description="Total flux of the detection in arbitrary units")
    instrumental_magnitude: float = Field(..., description="Instrumental magnitude of the detection")
    segmentation_index: int = Field(..., description="Index of the detection in the segmentation map")
    number_of_pixels: int = Field(..., description="Number of pixels in the detection")

    @classmethod
    def from_sep_object(cls, obj: dict) -> "Detection":
        return cls(
            x_centroid=obj["x"],
            y_centroid=obj["y"],
            semi_major_axis=obj["a"],
            semi_minor_axis=obj["b"],
            angle_to_horizon=obj["theta"],
            total_flux=obj["flux"],
            instrumental_magnitude=obj["inst_mag"],
            segmentation_index=obj["seg_idx"],
            number_of_pixels=obj["npix"],
        )

    @property
    def elongation(self) -> float:
        return self.semi_major_axis / self.semi_minor_axis

    def get_visual_magnitude(self, zero_point: float) -> float:
        return self.instrumental_magnitude + zero_point


class Detections(RootModel[list[Detection]]):

    @classmethod
    def from_sep_extract(cls, objects) -> "Detections":
        detections = [Detection.from_sep_object(obj) for obj in objects]
        return cls(detections)

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, index: int) -> Detection:
        return self.root[index]

    def __iter__(self):
        return iter(self.root)

    def append(self, detection: Detection) -> None:
        self.root.append(detection)

    @property
    def y_x_array(self) -> npt.NDArray[np.float32]:
        return np.array([[det.y_centroid, det.x_centroid] for det in self.root], dtype=np.float32)

    @property
    def x_y_array(self) -> npt.NDArray[np.float32]:
        return np.array([[det.x_centroid, det.y_centroid] for det in self.root], dtype=np.float32)

    @property
    def elongation_array(self) -> npt.NDArray[np.float32]:
        return np.array(
            [det.semi_major_axis / det.semi_minor_axis for det in self.root],
            dtype=np.float32,
        )

    @property
    def elongation_median(self) -> float:
        elongations = self.elongation_array
        return float(np.median(elongations))

    @property
    def elongation_sigma(self) -> float:
        elongations = self.elongation_array
        median = self.elongation_median
        mad = np.median(np.abs(elongations - median))
        return float(mad * MAD_TO_STD)


class SEPDetections:

    def __init__(self, objects) -> None:
        self._objects = objects
        self._objects.sort(order="flux")
        self._objects = self._objects[::-1]

    @property
    def count(self) -> int:
        return len(self._objects)

    @property
    def x_array(self) -> npt.NDArray[np.float32]:
        return self._objects["x"]

    @property
    def y_array(self) -> npt.NDArray[np.float32]:
        return self._objects["y"]

    @property
    def elongation_array(self) -> npt.NDArray[np.float32]:
        return self._objects["a"] / self._objects["b"]

    @property
    def elongation_median(self) -> float:
        elongations = self.elongation_array
        return float(np.median(elongations))

    @property
    def semi_major_axis_array(self) -> npt.NDArray[np.float32]:
        return self._objects["a"]

    @property
    def semi_minor_axis_array(self) -> npt.NDArray[np.float32]:
        return self._objects["b"]

    @property
    def pixel_count_array(self) -> npt.NDArray[np.int32]:
        return self._objects["npix"]

    @property
    def instrumental_magnitude_array(self) -> npt.NDArray[np.float32]:
        return self._objects["inst_mag"]

    @property
    def total_flux_array(self) -> npt.NDArray[np.float32]:
        return self._objects["flux"]

    @property
    def y_x_array(self) -> npt.NDArray[np.float32]:
        return np.vstack((self._objects["y"], self._objects["x"])).T



from pixelemon.optics._base_optical_assembly import BaseOpticalAssembly
from pixelemon.optics._arducam_m2508zh02 import ArducamM2508ZH02
from pixelemon.optics._arducam_cs2325zm01 import ArducamCS2325ZM01
from pixelemon.optics._arducam_cs2316zm02 import ArducamCS2316ZM02
from pixelemon.optics._williams_mini_cat_51 import WilliamsMiniCat51
from pixelemon.optics._goshyda_35 import Goshyda35

__all__ = [
    "BaseOpticalAssembly",
    "ArducamM2508ZH02",
    "ArducamCS2325ZM01",
    "ArducamCS2316ZM02",
    "WilliamsMiniCat51",
    "Goshyda35",
]



from math import ceil, pi, sqrt

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, Field

from pixelemon.constants import FWHM_TO_SIGMA
from pixelemon.logging import PixeLemonLog

MIN_DEBLEND_MESH_COUNT = 1
MAX_DEBLEND_MESH_COUNT = 64
DEFAULT_OPTICAL_AREA = 10.0
DEFAULT_ALBEDO = 0.3
DEFAULT_DETECTION_SIGMA = 2.0
DEFAULT_STAR_PIXEL_DISTANCE = 1.0
DEFAULT_SATELLITE_MAD_THRESHOLD = 3.0
DEFAULT_LIMITING_MAGNITUDE = 10.0
POINT_SOURCE_MIN_PIXEL_COUNT = 5
POINT_SOURCE_DEBLEND_MESH_COUNT = 32
POINT_SOURCE_DEBLEND_CONTRAST = 0.0005
POINT_SOURCE_FWHM = ceil(sqrt(POINT_SOURCE_MIN_PIXEL_COUNT / pi) * 2)
POINT_SOURCE_ARRAY_SIZE = int(6 * POINT_SOURCE_FWHM * FWHM_TO_SIGMA) | 1  # Ensure odd size
STREAK_SOURCE_MIN_PIXEL_COUNT = 5
STREAK_SOURCE_DEBLEND_MESH_COUNT = 1
STREAK_SOURCE_DEBLEND_CONTRAST = 1.0
STREAK_FWHM = 3
STREAK_SOURCE_ARRAY_SIZE = 65
DEFAULT_STAR_MAD_THRESHOLD = 1.5


class DetectionSettings(BaseModel):

    detection_sigma: float = Field(
        default=DEFAULT_DETECTION_SIGMA,
        description="The sigma threshold for detections",
    )

    star_elongation_threshold: float = Field(
        default=DEFAULT_STAR_MAD_THRESHOLD,
        description="The MAD threshold for star elongation boundary calculations",
        ge=0.1,
    )

    min_pixel_count: int = Field(
        ...,
        description="The minimum number of connected pixels to consider an object detected",
    )
    deblend_mesh_count: int = Field(
        ...,
        description="The number of meshes used for deblending overlapping objects",
        ge=MIN_DEBLEND_MESH_COUNT,
        le=MAX_DEBLEND_MESH_COUNT,
    )
    deblend_contrast: float = Field(
        ...,
        description="The contrast ratio used for deblending overlapping objects",
        ge=0.0,
        le=1.0,
    )
    merge_small_detections: bool = Field(
        ...,
        description="Whether to merge small detections with nearby larger ones",
    )

    kernel_array_size: int = Field(
        ...,
        description="The size of the Gaussian kernel array for detection",
    )

    full_width_half_maximum: int = Field(
        ...,
        description="The full width at half maximum (FWHM) for the Gaussian kernel",
    )

    satellite_area: float = Field(
        default=DEFAULT_OPTICAL_AREA,
        description="Optical area of the telescope in square meters",
        ge=0.01,
    )
    satellite_albedo: float = Field(
        default=DEFAULT_ALBEDO,
        description="Assumed albedo of the satellite for brightness estimation",
        ge=0.0,
        le=1.0,
    )
    limiting_magnitude: float = Field(
        default=DEFAULT_LIMITING_MAGNITUDE,
        description="Dimmest magnitude expected to be detected by the system",
    )
    satellite_elongation_threshold: float = Field(
        default=DEFAULT_SATELLITE_MAD_THRESHOLD,
        description="Elongation MAD multiple for which a detection is classified as a satellite",
        ge=1.0,
    )
    max_star_pixel_distance: float = Field(
        default=DEFAULT_STAR_PIXEL_DISTANCE,
        description="Maximum pixel distance between centroids to consider a source as a Tetra star",
        le=1.0,
    )
    satellite_pixel_threshold: float = Field(
        default=DEFAULT_SATELLITE_MAD_THRESHOLD,
        description="Pixel count MAD multiple for which a detection is classified as a satellite",
        ge=1.0,
    )

    def model_post_init(self, _context):
        PixeLemonLog().info(f"Detection detection sigma set to {self.detection_sigma}")
        PixeLemonLog().info(f"Detection minimum pixel count set to {self.min_pixel_count}")
        PixeLemonLog().info(f"Detection deblend mesh count set to {self.deblend_mesh_count}")
        PixeLemonLog().info(f"Detection deblend contrast set to {self.deblend_contrast}")
        PixeLemonLog().info(f"Detection merge small detections set to {self.merge_small_detections}")
        PixeLemonLog().info(f"Detection kernel array size set to {self.kernel_array_size}")
        PixeLemonLog().info(f"Detection FWHM set to {self.full_width_half_maximum}")
        PixeLemonLog().info(f"Detection satellite optical area set to {self.satellite_area} m^2")
        PixeLemonLog().info(f"Detection satellite albedo set to {self.satellite_albedo}")
        PixeLemonLog().info(f"Detection limiting magnitude set to {self.limiting_magnitude}")
        PixeLemonLog().info(f"Detection satellite elongation threshold set to {self.satellite_elongation_threshold}")
        PixeLemonLog().info(f"Detection star elongation threshold set to {self.star_elongation_threshold}")
        PixeLemonLog().info(f"Detection satellite pixel threshold set to {self.satellite_pixel_threshold}")

    @classmethod
    def point_source_defaults(cls) -> "DetectionSettings":

        return cls(
            min_pixel_count=POINT_SOURCE_MIN_PIXEL_COUNT,
            deblend_mesh_count=POINT_SOURCE_DEBLEND_MESH_COUNT,
            deblend_contrast=POINT_SOURCE_DEBLEND_CONTRAST,
            merge_small_detections=False,
            kernel_array_size=POINT_SOURCE_ARRAY_SIZE,
            full_width_half_maximum=POINT_SOURCE_FWHM,
        )

    @classmethod
    def streak_source_defaults(cls) -> "DetectionSettings":
        return cls(
            min_pixel_count=STREAK_SOURCE_MIN_PIXEL_COUNT,
            deblend_mesh_count=STREAK_SOURCE_DEBLEND_MESH_COUNT,
            deblend_contrast=STREAK_SOURCE_DEBLEND_CONTRAST,
            merge_small_detections=True,
            kernel_array_size=STREAK_SOURCE_ARRAY_SIZE,
            full_width_half_maximum=STREAK_FWHM,
        )

    @property
    def gaussian_kernel(self) -> npt.NDArray[np.float32]:
        ax = np.array(self.kernel_array_size) - self.kernel_array_size // 2
        xx, yy = np.meshgrid(ax, ax)
        sigma = self.full_width_half_maximum * FWHM_TO_SIGMA
        kernel = np.exp(-(xx**2 + yy**2) / (2.0 * sigma**2))
        return kernel / kernel.sum()



from pydantic import BaseModel, Field

DEFAULT_ANGLE_DISTANCE_CORRELATION = 0.1  # in degrees


class CorrelationSettings(BaseModel):

    centroid_angle_limit: float = Field(
        default=DEFAULT_ANGLE_DISTANCE_CORRELATION,
        description="Maximum allowed centroid angle difference for correlation in degrees",
        ge=0.0,
        le=180.0,
    )



from pixelemon.processing._background_settings import BackgroundSettings, MIN_BACKGROUND_MESH_COUNT
from pixelemon.processing._detection_settings import DetectionSettings
from pixelemon.processing._detection import Detection, Detections, DetectionLine, SEPDetections
from pixelemon.processing._observation import PixeLemonObservation, PixeLemonObservations

__all__ = [
    "BackgroundSettings",
    "MIN_BACKGROUND_MESH_COUNT",
    "DetectionSettings",
    "Detection",
    "Detections",
    "DetectionLine",
    "SEPDetections",
    "PixeLemonObservation",
    "PixeLemonObservations",
]



from pydantic import Field

from pixelemon.optics._base_optical_assembly import BaseOpticalAssembly


class WilliamsMiniCat51(BaseOpticalAssembly):
    image_circle_diameter: float = Field(default=43.2, frozen=True)
    focal_length: float = Field(default=178.0, frozen=True)
    focal_ratio: float = Field(default=3.5, frozen=True)



import logging
import sys
import warnings

from astropy.utils.exceptions import AstropyUserWarning, AstropyWarning  # type: ignore
from astropy.wcs import FITSFixedWarning  # type: ignore

COLORS = {
    "DEBUG": "\033[36m",  # cyan
    "INFO": "\033[32m",  # green
    "WARNING": "\033[33m",  # yellow
    "ERROR": "\033[31m",  # red
    "CRITICAL": "\033[1;31m",  # bold red
}
RESET = "\033[0m"

logging.basicConfig(
    level=logging.CRITICAL,
    force=True,  # Python 3.8+: replace existing handlers
)

logging.getLogger("astropy").setLevel(logging.CRITICAL + 1)
logging.getLogger("astropy").disabled = True

warnings.filterwarnings("ignore", category=AstropyWarning)
warnings.filterwarnings("ignore", category=AstropyUserWarning)
warnings.filterwarnings("ignore", category=FITSFixedWarning)
warnings.filterwarnings(
    "ignore", message=".*WCS.all_world2pix.*failed to converge.*", category=UserWarning, module="astropy"
)


class LevelColorFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        color = COLORS.get(record.levelname, "")
        levelname_colored = f"{color}{record.levelname}{RESET}"
        record.levelname = levelname_colored
        return super().format(record)


class PixeLemonLog(logging.Logger):
    _instance: "PixeLemonLog | None" = None
    _initialized: bool = False

    def __new__(cls, *args, **kwargs) -> "PixeLemonLog":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if not self._initialized:
            pkg = __name__.split(".")[0]
            super().__init__(pkg)
            self.setLevel(logging.INFO)
            self.propagate = False
            if not self.handlers:
                h = logging.StreamHandler(sys.stdout)
                fmt = "%(asctime)s.%(msecs)03d: [%(levelname)s] %(message)s"
                datefmt = "%Y-%m-%d %H:%M:%S"
                h.setFormatter(LevelColorFormatter(fmt=fmt, datefmt=datefmt))
                self.addHandler(h)
            self._initialized = True

    def set_level(self, level: int) -> None:
        self.setLevel(level)
        for handler in self.handlers:
            handler.setLevel(level)



from pydantic import BaseModel


class CorrelatedDetection(BaseModel):
    satellite_id: str
    right_ascension: float
    declination: float
    visual_magnitude: float
    angle_between_centroids: float
    cross_line_of_sight_range: float



from datetime import datetime
from pathlib import Path
from uuid import UUID

import numpy as np
import numpy.typing as npt
import plotly.express as px  # type: ignore
import plotly.graph_objects as go  # type: ignore
import sep  # type: ignore
from astropy.coordinates import SkyCoord  # type: ignore
from astropy.io import fits  # type: ignore
from astropy.visualization import AsinhStretch, ZScaleInterval  # type: ignore
from astropy.wcs import WCS  # type: ignore
from astropy.wcs.utils import fit_wcs_from_points  # type: ignore
from keplemon.bodies import Constellation, Observatory
from keplemon.elements import TopocentricElements
from keplemon.enums import ReferenceFrame
from keplemon.time import Epoch, TimeSpan
from numpy.lib import recfunctions as rfn
from pydantic import BaseModel, Field

from pixelemon._plate_solve import PlateSolve, TetraSolver
from pixelemon._telescope import Telescope
from pixelemon.constants import BASE_TO_KILO, KILO_TO_BASE, PERCENT_TO_DECIMAL, SUN_MAGNITUDE
from pixelemon.correlation import CorrelatedDetection, CorrelationCandidate, CorrelationCandidates, CorrelationSettings
from pixelemon.logging import PixeLemonLog
from pixelemon.processing import (
    MIN_BACKGROUND_MESH_COUNT,
    BackgroundSettings,
    Detection,
    DetectionSettings,
    PixeLemonObservation,
    PixeLemonObservations,
    SEPDetections,
)
from pixelemon.utils import map_outside_std_threshold, map_within_std_threshold


class TelescopeImage(BaseModel):
    _original_array: npt.NDArray[np.float32] | None = None
    _processed_array: npt.NDArray[np.float32] | None = None
    _plate_solve: PlateSolve | None = None
    _detections: SEPDetections | None = None
    _background: sep.Background | None = None
    _background_removed: bool = False
    epoch: datetime | None = None
    telescope: Telescope | None = None
    exposure_time: float | None = None
    _ground_site: Observatory | None = None
    _wcs: WCS | None = None
    image_scale: float = Field(default=1.0, description="The image scale due to cropping")
    background_settings: BackgroundSettings = Field(default=BackgroundSettings())
    detection_settings: DetectionSettings = Field(default=DetectionSettings.streak_source_defaults())
    correlation_settings: CorrelationSettings = Field(default=CorrelationSettings())

    @classmethod
    def from_fits_file(cls, file_path: Path, telescope: Telescope) -> "TelescopeImage":
        with fits.open(file_path) as hdul:
            img = cls()
            assert hasattr(hdul[0], "header")
            header = getattr(hdul[0], "header")
            img.exposure_time = header["EXPTIME"]
            img._wcs = WCS(header)
            img.epoch = datetime.fromisoformat(header["DATE-OBS"])
            img.telescope = telescope
            img._original_array = getattr(hdul[0], "data").astype(np.float32)

            if "SITELAT" in header and "SITELONG" in header and "SITEALT" in header:
                lat = header["SITELAT"]
                lon = header["SITELONG"]
                h = header["SITEALT"] * BASE_TO_KILO
            elif "OBSGEO-L" in header and "OBSGEO-B" in header and "OBSGEO-H" in header:
                lon = header["OBSGEO-L"]
                lat = header["OBSGEO-B"]
                h = header["OBSGEO-H"] * BASE_TO_KILO
            else:
                site_msg = "(SITELAT, SITELONG, and SITEALT)"
                geo_msg = "(OBSGEO-L, OBSGEO-B, and OBSGEO-H)"
                raise ValueError(f"Location not found in FITS header. Expected either {site_msg} or {geo_msg}.")

            img._ground_site = Observatory(lat, lon, h)
            assert img._original_array is not None
            actual_ratio = img._original_array.shape[1] / img._original_array.shape[0]
            if not np.isclose(img.telescope.aspect_ratio, actual_ratio, atol=1e-6):
                new_width = int(img._original_array.shape[0] * img.telescope.aspect_ratio)
                img._original_array = img._original_array[:, 0:new_width]
                img._original_array = np.ascontiguousarray(img._original_array)
                new_shape = img._original_array.shape
                PixeLemonLog().debug(f"Trimmed {file_path} to expected {new_shape[0]}h x {new_shape[1]}w")
            assert img._original_array is not None
            img._processed_array = img._original_array.copy()
        return img

    def get_correlation_candidates(self, sats: Constellation) -> CorrelationCandidates:
        if self._ground_site is None:
            raise ValueError("Ground site is not set.")
        solve = self.plate_solve
        if solve is None:
            return CorrelationCandidates([])
        elif self.epoch is None:
            raise ValueError("Image epoch is not set.")
        elif self.exposure_time is None:
            raise ValueError("Exposure time is not set.")

        kepoch = Epoch.from_datetime(self.epoch) + TimeSpan.from_seconds(self.exposure_time * 0.5)

        sats_in_fov = self._ground_site.get_field_of_view_report(
            kepoch,
            TopocentricElements.from_j2000(
                kepoch,
                solve.right_ascension,
                solve.declination,
            ),
            self.horizontal_field_of_view * 0.5,
            sats,
            ReferenceFrame.J2000,
        )

        correlation_candidates = CorrelationCandidates([])
        for start_candidate in sats_in_fov.candidates:

            r = start_candidate.direction.range
            assert r is not None
            r = r * KILO_TO_BASE
            area = self.detection_settings.satellite_area
            albedo = self.detection_settings.satellite_albedo
            vis_mag = SUN_MAGNITUDE + 2.5 * np.log10((4 * np.pi * r**2) / (albedo * area))
            if vis_mag <= self.detection_settings.limiting_magnitude:
                sat_range = start_candidate.direction.range
                assert sat_range is not None

                xc, yc = self.get_fits_pixels(
                    start_candidate.direction.right_ascension,
                    start_candidate.direction.declination,
                )

                correlation_candidate = CorrelationCandidate(
                    id=start_candidate.satellite_id,
                    x_centroid=xc,
                    y_centroid=yc,
                    range=sat_range,
                )
                correlation_candidates.root.append(correlation_candidate)

        return correlation_candidates

    def get_angular_separation(self, x0: float, y0: float, x1: float, y1: float) -> float:
        if self._wcs is None:
            raise ValueError("WCS is not set.")
        sky0 = self._wcs.pixel_to_world(x0, y0)
        sky1 = self._wcs.pixel_to_world(x1, y1)
        return sky0.separation(sky1).deg

    @property
    def star_elongation_histogram(self) -> go.Histogram:
        return go.Histogram(
            x=self.stars.elongation_array,
            xbins=dict(
                start=self.detections.elongation_array.min(),
                end=self.detections.elongation_array.max(),
                size=np.std(self.detections.elongation_array) / 10,
            ),
            name="Star Elongations",
            marker_color="yellow",
            autobinx=False,
            opacity=0.75,
        )

    @property
    def star_pixel_count_histogram(self) -> go.Histogram:
        return go.Histogram(
            x=self.stars.pixel_count_array,
            autobinx=False,
            xbins=dict(
                start=self.detections.pixel_count_array.min(),
                end=self.detections.pixel_count_array.max(),
                size=np.std(self.detections.pixel_count_array) / 10,
            ),
            name="Star Pixel Count Histogram",
            marker_color="yellow",
            opacity=0.75,
        )

    @property
    def satellite_elongation_histogram(self) -> go.Histogram:
        return go.Histogram(
            x=self.satellites.elongation_array,
            autobinx=False,
            xbins=dict(
                start=self.detections.elongation_array.min(),
                end=self.detections.elongation_array.max(),
                size=np.std(self.detections.elongation_array) / 10,
            ),
            name="Satellite Elongations",
            marker_color="red",
            opacity=0.75,
        )

    @property
    def satellite_pixel_count_histogram(self) -> go.Histogram:
        return go.Histogram(
            x=self.satellites.pixel_count_array,
            autobinx=False,
            xbins=dict(
                start=self.detections.pixel_count_array.min(),
                end=self.detections.pixel_count_array.max(),
                size=np.std(self.detections.pixel_count_array) / 10,
            ),
            name="Satellite Pixel Count Histogram",
            marker_color="red",
            opacity=0.75,
        )

    @property
    def elongation_histogram(self) -> go.Histogram:
        return go.Histogram(
            x=self.detections.elongation_array,
            autobinx=False,
            xbins=dict(
                start=self.detections.elongation_array.min(),
                end=self.detections.elongation_array.max(),
                size=np.std(self.detections.elongation_array) / 10,
            ),
            name="All Detections",
            marker_color="gray",
            opacity=0.75,
        )

    @property
    def pixel_count_histogram(self) -> go.Histogram:
        return go.Histogram(
            x=self.detections.pixel_count_array,
            autobinx=False,
            xbins=dict(
                start=self.detections.pixel_count_array.min(),
                end=self.detections.pixel_count_array.max(),
                size=np.std(self.detections.pixel_count_array) / 10,
            ),
            name="All Detections",
            marker_color="gray",
            opacity=0.75,
        )

    def plot_elongation_histogram(self):
        fig = go.Figure()
        fig.add_trace(self.elongation_histogram)
        fig.add_trace(self.star_elongation_histogram)
        fig.add_trace(self.satellite_elongation_histogram)
        fig.show(config={"staticPlot": False})

    def plot_pixel_count_histogram(self):
        fig = go.Figure()
        fig.add_trace(self.pixel_count_histogram)
        fig.add_trace(self.star_pixel_count_histogram)
        fig.add_trace(self.satellite_pixel_count_histogram)
        fig.show(config={"staticPlot": False})

    @property
    def satellites_scatter(self) -> go.Scatter:
        yx = self.satellites.y_x_array
        xs, ys = yx[:, 1], self.height - 1 - yx[:, 0]
        return go.Scatter(
            x=xs, y=ys, mode="markers", marker=dict(color="red", size=5, symbol="x"), name="Possible Satellites"
        )

    @property
    def stars_scatter(self) -> go.Scatter:
        yx = self.stars.y_x_array
        xs, ys = yx[:, 1], self.height - 1 - yx[:, 0]
        return go.Scatter(
            x=xs, y=ys, mode="markers", marker=dict(color="yellow", size=5, symbol="circle"), name="Possible Stars"
        )

    @property
    def matched_stars_scatter(self) -> go.Scatter:
        if self.plate_solve is None:
            raise ValueError("Plate solve is not set.")
        yx = self.plate_solve.matched_centroids
        x, y = yx[:, 1], self.height - 1 - yx[:, 0]
        return go.Scatter(
            x=x, y=y, mode="markers", marker=dict(color="green", size=7, symbol="star"), name="Matched Stars"
        )

    def get_candidates_scatter(self, sats: Constellation) -> go.Scatter:
        candidates = self.get_correlation_candidates(sats)
        xy = np.array([(c.x_centroid, c.y_centroid) for c in candidates.root], dtype=np.float64)
        labels = [sats[c.id].name for c in candidates.root]
        return go.Scatter(
            x=xy[:, 0],
            y=xy[:, 1],
            mode="markers",
            marker=dict(color="blue", size=5, symbol="diamond"),
            name="Candidates",
            text=labels,
            textposition="top center",
        )

    def plot(self, sats: Constellation):

        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        elif self.epoch is None:
            raise ValueError("Image epoch is not set.")

        interval = ZScaleInterval()
        vmin, vmax = interval.get_limits(self._processed_array)
        stretch = AsinhStretch()
        data_stretched = stretch((np.clip(self._processed_array, vmin, vmax) - vmin) / (vmax - vmin))
        data = np.flipud(data_stretched)

        fig = px.imshow(data, color_continuous_scale="Viridis", origin="lower", title=f"{self.epoch.isoformat()} Image")

        fig.add_trace(self.satellites_scatter)
        fig.add_trace(self.stars_scatter)
        fig.add_trace(self.matched_stars_scatter)
        fig.add_trace(self.get_candidates_scatter(sats))
        fig.update_traces(showscale=False, selector=dict(type="heatmap"))
        fig.show(config={"staticPlot": False})

    def get_angles(self, x: float, y: float) -> tuple[float, float]:
        if self._wcs is None:
            raise ValueError("WCS is not set.")
        sky = self._wcs.pixel_to_world(x, y)
        return sky.ra.deg, sky.dec.deg

    def get_correlation_to_candidate(self, candidate: CorrelationCandidate) -> CorrelatedDetection | None:
        detections = self.satellites

        det_x, det_y = detections.x_array, self.height - 1 - detections.y_array
        delta_pixels = candidate.x_centroid - det_x, candidate.y_centroid - det_y
        distances = np.hypot(delta_pixels[0], delta_pixels[1])
        assert self.telescope is not None
        pixel_to_degree = self.telescope.diagonal_pixel_scale
        angular_distances = distances * pixel_to_degree
        valid_angles = angular_distances[angular_distances <= self.correlation_settings.centroid_angle_limit]

        if len(valid_angles) == 0:
            PixeLemonLog().debug("No detections within centroid angle limit")
            return None
        else:
            best_i = int(np.argmin(angular_distances))

        y, x = detections.y_x_array[best_i]
        ra_dec = self.get_angles(x, y)
        zp = self.zero_point
        assert zp is not None
        return CorrelatedDetection(
            satellite_id=candidate.id,
            right_ascension=ra_dec[0],
            declination=ra_dec[1],
            visual_magnitude=detections.instrumental_magnitude_array[best_i] + zp,
            angle_between_centroids=angular_distances[best_i],
            cross_line_of_sight_range=candidate.range * np.deg2rad(angular_distances[best_i]),
        )

    def get_correlations(self, sats: Constellation) -> list[CorrelatedDetection]:
        detections = self.satellites
        candidates = self.get_correlation_candidates(sats)
        if self._wcs is None:
            raise ValueError("WCS is not set.")
        elif len(candidates) == 0:
            PixeLemonLog().info("No correlation candidates in field of view")
            return []
        elif detections.count == 0:
            PixeLemonLog().info("No satellites detected in image")
            return []

        # Pixels -> sky in one shot
        det_xy = np.column_stack((detections.x_array, self.height - 1 - detections.y_array))
        sat_xy = np.array([(c.x_centroid, c.y_centroid) for c in candidates], dtype=np.float64)
        det_sky = self._wcs.pixel_to_world(det_xy[:, 0], det_xy[:, 1])
        sat_sky = self._wcs.pixel_to_world(sat_xy[:, 0], sat_xy[:, 1])

        # Pairwise angular distances (deg)
        distances = det_sky[:, None].separation(sat_sky[None, :]).deg
        limit = self.correlation_settings.centroid_angle_limit
        distances = np.where(distances <= limit, distances, np.inf)

        order = np.argsort(distances, axis=None)
        det_used = np.zeros(detections.count, dtype=bool)
        sat_used = np.zeros(len(candidates), dtype=bool)
        correlations: list[CorrelatedDetection] = []
        zp = self.zero_point
        if zp is None:
            return correlations

        for flat_idx in order:
            det_idx = flat_idx // len(candidates)
            sat_idx = flat_idx % len(candidates)
            angle = float(distances[det_idx, sat_idx])
            if not np.isfinite(angle):
                break
            if det_used[det_idx] or sat_used[sat_idx]:
                continue

            det_used[det_idx] = True
            sat_used[sat_idx] = True

            cand = candidates[sat_idx]
            ra, dec = det_sky[det_idx].ra.deg, det_sky[det_idx].dec.deg
            correlations.append(
                CorrelatedDetection(
                    satellite_id=cand.id,
                    right_ascension=ra,
                    declination=dec,
                    visual_magnitude=float(detections.instrumental_magnitude_array[det_idx]) + zp,
                    angle_between_centroids=angle,
                    cross_line_of_sight_range=cand.range * np.deg2rad(angle),
                )
            )

        PixeLemonLog().info(f"Correlated {len(correlations)} of {self.satellites.count} detections")

        return correlations

    def get_observations(self, sats: Constellation) -> PixeLemonObservations:
        correlations = self.get_correlations(sats)
        observations = []
        if self.epoch is None:
            raise ValueError("Image epoch is not set.")
        elif self._ground_site is None:
            raise ValueError("Ground site is not set.")
        elif self.telescope is None:
            raise ValueError("Telescope is not set.")

        for corr in correlations:
            observation = PixeLemonObservation(
                satellite_id=UUID(corr.satellite_id),
                epoch=self.epoch,
                right_ascension=corr.right_ascension,
                declination=corr.declination,
                sensor_latitude=self._ground_site.latitude,
                sensor_longitude=self._ground_site.longitude,
                sensor_altitude=(self._ground_site.altitude * KILO_TO_BASE),
                angular_noise=self.telescope.diagonal_pixel_scale,
                visual_magnitude=corr.visual_magnitude,
            )
            observations.append(observation)
        PixeLemonLog().info(f"Generated {len(observations)} observations")
        return PixeLemonObservations(observations)

    def get_correlation_to_detection(self, det: Detection, sats: CorrelationCandidates) -> CorrelatedDetection | None:

        xc, yc = det.x_centroid, self.height - 1 - det.y_centroid
        angular_distances = []

        for candidate in sats:
            angle = self.get_angular_separation(xc, yc, candidate.x_centroid, candidate.y_centroid)
            if angle < self.correlation_settings.centroid_angle_limit:
                angular_distances.append(angle)

        if not angular_distances:
            PixeLemonLog().debug("No candidates within centroid angle limit")
            return None
        elif len(angular_distances) == 1:
            best_i = 0
        else:
            best_i = int(np.argmin(angular_distances))

        candidate = sats[best_i]
        ra_dec = self.get_angles(det.x_centroid, det.y_centroid)
        zp = self.zero_point
        assert zp is not None
        return CorrelatedDetection(
            satellite_id=candidate.id,
            right_ascension=ra_dec[0],
            declination=ra_dec[1],
            visual_magnitude=det.get_visual_magnitude(zp),
            angle_between_centroids=angular_distances[best_i],
            cross_line_of_sight_range=candidate.range * np.deg2rad(angular_distances[best_i]),
        )

    def write_to_fits_file(self, file_path: Path):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        fits.writeto(file_path, self._processed_array.astype("uint8"), overwrite=True)

    def dampen_hot_rows(self):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        row_means = np.mean(self._processed_array, axis=1)
        global_mean = np.mean(row_means)
        global_std = np.std(row_means)
        threshold = global_mean + self.detection_settings.detection_sigma * global_std
        for i, row_mean in enumerate(row_means):
            if row_mean > threshold:
                PixeLemonLog().debug(f"Dampening hot row {i} with mean {row_mean:.2f}")
                self._processed_array[i, :] *= global_mean / row_mean

    def crop(self, crop_percent: float):
        if self._original_array is None:
            raise ValueError("Image array is not loaded.")

        crop_fraction = crop_percent * PERCENT_TO_DECIMAL
        height, width = self._original_array.shape
        crop_height = int(height * crop_fraction / 2)
        crop_width = int(width * crop_fraction / 2)
        self._processed_array = np.ascontiguousarray(
            self._original_array[crop_height : height - crop_height, crop_width : width - crop_width]
        )
        self.image_scale = self.image_scale * (1.0 - crop_fraction)
        self._plate_solve = None
        self._satellites = None
        self._segmentation_map = None
        self._star_segments = None
        self._stars = None
        self._background = None
        self._background_removed = False

    @property
    def background(self) -> sep.Background:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        if self._background is None:

            self.dampen_hot_rows()

            bw = max(
                MIN_BACKGROUND_MESH_COUNT, int(self._processed_array.shape[1] / self.background_settings.mesh_count)
            )
            bh = max(
                MIN_BACKGROUND_MESH_COUNT, int(self._processed_array.shape[0] / self.background_settings.mesh_count)
            )

            self._background = sep.Background(
                self._processed_array,
                bw=bw,
                bh=bh,
                fw=self.background_settings.filter_size,
                fh=self.background_settings.filter_size,
                fthresh=self.background_settings.detection_threshold,
            )

        return self._background

    def remove_background(self):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        self._processed_array = self._processed_array - self.background
        self._background_removed = True

    def reset(self):
        if self._original_array is None:
            raise ValueError("Image array is not loaded.")
        self._processed_array = self._original_array.copy()
        self.image_scale = 1.0
        self._background = None
        self._detections = None
        self._plate_solve = None
        self._background_removed = False

    def get_fits_pixels(self, ra: float, dec: float) -> tuple[float, float]:
        if self._wcs is None:
            raise ValueError("WCS is not set.")
        sky = SkyCoord(ra, dec, unit="deg")
        x, y = self._wcs.world_to_pixel(sky)
        return x, y

    @property
    def horizontal_field_of_view(self) -> float:
        if self.telescope is None:
            raise ValueError("Telescope is not set.")
        return self.telescope.horizontal_field_of_view * self.image_scale

    @property
    def vertical_field_of_view(self) -> float:
        if self.telescope is None:
            raise ValueError("Telescope is not set.")
        return self.telescope.vertical_field_of_view * self.image_scale

    @property
    def diagonal_field_of_view(self) -> float:
        if self.telescope is None:
            raise ValueError("Telescope is not set.")
        return self.telescope.diagonal_field_of_view * self.image_scale

    @property
    def detections(self) -> SEPDetections:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        if not self._background_removed:
            self.remove_background()

        if self._detections is None:
            objects = sep.extract(
                self._processed_array,
                thresh=self.detection_settings.detection_sigma * self.background.globalrms,
                minarea=self.detection_settings.min_pixel_count,
                filter_kernel=self.detection_settings.gaussian_kernel,
                deblend_nthresh=self.detection_settings.deblend_mesh_count,
                deblend_cont=self.detection_settings.deblend_contrast,
                clean=self.detection_settings.merge_small_detections,
                segmentation_map=False,
            )

            # add instrumental magnitude to the objects
            instrumental_mag = -2.5 * np.log10(objects["flux"] / self.exposure_time)
            objects = rfn.append_fields(objects, "inst_mag", instrumental_mag, usemask=False)

            self._detections = SEPDetections(objects)

        return self._detections

    @property
    def stars(self) -> SEPDetections:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        PixeLemonLog().debug(f"{self.detections.count} total detections before star filtering")
        elongation_map = map_within_std_threshold(
            self.detections.elongation_array,
            self.detection_settings.star_elongation_threshold,
            use_mad=True,
        )
        PixeLemonLog().debug(f"{len(elongation_map[elongation_map])} star-like objects passed elongation filter")
        return SEPDetections(self.detections._objects[elongation_map])

    @property
    def satellites(self):
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")

        PixeLemonLog().debug(f"{self.detections.count} total detections before satellite filtering")
        elongation_map = map_outside_std_threshold(
            self.detections.elongation_array,
            self.detection_settings.satellite_elongation_threshold,
            use_mad=True,
        )
        PixeLemonLog().debug(f"{len(elongation_map[elongation_map])} satellite-like objects passed elongation filter")
        return SEPDetections(self.detections._objects[elongation_map])

    @property
    def plate_solve(self) -> PlateSolve | None:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        if self.telescope is None:
            raise ValueError("Telescope is not set.")

        if self._plate_solve is None:

            plate_solve = TetraSolver().solve_from_centroids(
                self.stars.y_x_array,
                img_size=(self.height, self.width),
                fov=self.diagonal_field_of_view,
            )

            if plate_solve is not None:

                pixel_scale = self.telescope.horizontal_pixel_scale
                assert self._wcs is not None

                # seed the WCS to improve chances of solution with fit_wcs_from_points
                self._wcs.wcs.crpix = [self._processed_array.shape[1] / 2, self._processed_array.shape[0] / 2]
                self._wcs.wcs.crval = [plate_solve.right_ascension, plate_solve.declination]
                self._wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
                self._wcs.wcs.cunit = ["deg", "deg"]
                theta = np.deg2rad(-plate_solve.roll)
                cd11 = -pixel_scale * np.cos(theta)
                cd12 = pixel_scale * np.sin(theta)
                cd21 = pixel_scale * np.sin(theta)
                cd22 = pixel_scale * np.cos(theta)
                self._wcs.wcs.cd = np.array([[cd11, cd12], [cd21, cd22]], dtype=float)

                # invert matched centroids for WCS fitting
                yx = plate_solve.matched_centroids
                x, y = yx[:, 1], self.height - 1 - yx[:, 0]
                ra_dec = plate_solve.matched_stars

                # fit WCS from matched stars
                sky = SkyCoord(ra=ra_dec[:, 0], dec=ra_dec[:, 1], unit="deg")
                self._wcs = fit_wcs_from_points((x, y), sky, sip_degree=5, proj_point="center")
                self._plate_solve = plate_solve
                PixeLemonLog().info(f"Matched {plate_solve.number_of_stars} of {self.stars.count} stars")

        return self._plate_solve

    @property
    def height(self) -> int:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        return self._processed_array.shape[0]

    @property
    def width(self) -> int:
        if self._processed_array is None:
            raise ValueError("Image array is not loaded.")
        return self._processed_array.shape[1]

    @property
    def zero_point(self) -> float | None:
        solve = self.plate_solve
        if solve is None:
            return None

        PixeLemonLog().debug(f"Calculating zero point from {solve.matched_centroids.shape[0]} matched stars")
        tetra_yx = solve.matched_centroids
        stars_yx = self.stars.y_x_array
        inst_mag = self.stars.instrumental_magnitude_array

        delta = stars_yx[:, None, :] - tetra_yx[None, :, :]
        distance = np.linalg.norm(delta, axis=2)

        closest_star_idx = np.argmin(distance, axis=0)
        closest_dist = distance[closest_star_idx, np.arange(distance.shape[1])]
        match_mask = closest_dist <= self.detection_settings.max_star_pixel_distance

        if not match_mask.any():
            return None

        matched_star_idx = closest_star_idx[match_mask]
        matched_tetra_idx = np.arange(tetra_yx.shape[0])[match_mask]

        offsets = solve.matched_stars[matched_tetra_idx, 2] - inst_mag[matched_star_idx]
        PixeLemonLog().debug(f"Zero point calculated from {len(offsets)} matched stars")
        PixeLemonLog().debug(f"Max distance for matched stars: {closest_dist[match_mask].max():.12f} pixels")
        return float(offsets.mean())



from pixelemon.correlation._candidates import CorrelationCandidates, CorrelationCandidate
from pixelemon.correlation._correlation_settings import CorrelationSettings
from pixelemon.correlation._correlated_detection import CorrelatedDetection


__all__ = ["CorrelationCandidate", "CorrelationCandidates", "CorrelationSettings", "CorrelatedDetection"]
